# Author Aimped
# Date: 2023-May-25
# Description: This file contains standart input and output (sio) keys for each task

import json
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Extra, Field

data_types = [
    'data_json',
    'data_pdf',
    'data_image',
    'data_svg',
    'data_audio',
    'data_csv',
    'data_txt',
    'data_excel',
    'data_docx',
    'data_xml',
    'data_video',
    'data_zip',
    'data_char',
    'data_file',
    'raw_output',
    'data_url'
    'data_url_speak',
    'data_audio_speak',
]


################ Task: Text Classification ################

class TextClassificationInput(BaseModel):
    """Input data for text classification task"""
    text: List[str] = Field(..., description="List of texts", example=["This is an example text for classification.",
                                                                       "This is another example text for classification."])
    class Config:
        extra = Extra.allow

class TextClassificationOutput(BaseModel):
    """Output data for text classification task"""
    status: bool = True
    data_type: List[str] = Field(default=["data_json"], description="List of data types", enum=data_types)
    output: Dict[str, Any] = Field(
        default={"data_json": {"result": None}},
        description="data_json output dictionary",
        keys={"enum": data_types},
        example={
            'data_json': {
                'result': [
                    {
                        'category': [str],
                        'classes': [{'label': str, 'score': float},
                                    {'label': str, 'score': float}]
                    }
                ]
            }
        }
    )

    def __init__(self, model_prediction: Any, **data: Any):
        super().__init__(**data)
        self.output['data_json']['result'] = model_prediction


################ Task: Name Entity Recognition ################

class NameEntityRecognitionInput(BaseModel):
    """Input data for name entity recognition task"""
    text: List[str] = Field(..., description="List of texts", example=["This is an example text for NER."])
    entity: List[Any] = Field(..., description="List of entities", example=["PERSON", "ORG", "GPE"])

    class Config:
        extra = Extra.allow


class NameEntityRecognitionOutput(BaseModel):
    """Output data for name entity recognition task"""
    status: bool = True
    data_type: List[str] = Field(default=["data_json"], description="List of data types", enum=data_types)
    output: Dict[str, Any] = Field(
        default={"data_json": {"result": None}},
        description="Output dictionary",
        keys={"enum": data_types},
        example={
            'data_json': {
                'result': [
                    [{
                        'entity': str,
                        'confidence': float,
                        'chunk': str,
                        'begin': int,
                        'end': int
                    }]
                ]
            }
        }
    )

    def __init__(self, model_prediction: Any, **data: Any):
        super().__init__(**data)
        self.output['data_json']['result'] = model_prediction


######################### Task: Deidentification #########################

class DeidentificationInput(BaseModel):
    """Input data for deidentification task"""
    text: List[str] = Field(..., description="List of texts", example=["This is an example text for deidentification."])
    entity: List[str] = Field(..., description="List of entities", example=["PERSON", "ORG", "GPE"])

    class Config:
        extra = Extra.allow

class DeidentificationOutput(BaseModel):
    """Output data for deidentification task"""
    status: bool = True
    data_type: List[str] = Field(default=["data_json"], description="List of data types", enum=data_types)
    output: Dict[str, Any] = Field(
                                    default={"data_json": {"result": None}},
                                    description="Output dictionary",
                                    keys={"enum": data_types},
                                    example={
                                        'data_json': {'result': [
                                            {'entities': [
                                                {
                                                    'entity': str,
                                                    'confidence': float,
                                                    'chunk': str,
                                                    'begin': int,
                                                    'end': int,
                                                    'faked_chunk': str
                                                }],
                                                'masked_text': str,
                                                'faked_text': str
                                            }]
                                        }}
                                )

    def __init__(self, model_prediction: Any, **data: Any):
        super().__init__(**data)
        self.output['data_json']['result'] = model_prediction


####################### Task: Chatbot ###################### TODO: Still in development

class ChatbotInput(BaseModel):
    """Input data for chatbot task"""
    messages: List[Dict[str, str]] = Field(..., description="List of messages",
                                           example=[{"role": "user", "content": "Who are you?"},
                                                    {"role": "assistant", "content": "I am Chat Bot. Who are you?"},
                                                    {"role": "user", "content": "I am Joseph."}
                                                    ])
    stream: bool = False
    max_tokens: int = 256
    temperature: float = 0.7
    data_files: Dict[str, Union[List[str], None]] = Field(default=None, description="Data files dictionary including text, audio, image, pdf, urls in list", example={"data_txt": ["https://cdn-dev.aimped.ai/temporary/2023-08-28-Global_Economic_Landscape.txt"]})

    class Config:
        extra = Extra.allow


class ChatbotOutput(BaseModel):
    """Output data for chatbot task"""
    status: bool = True
    data_type: List[str] = Field(default_factory=lambda: ["data_json", "data_audio_speak", "data_url_speak"],
                                 description="List of data types",
                                 enum=data_types)

    output: Dict[str, Any] = Field(default_factory=lambda: dict(),
        description="Output dictionary",
        keys={"enum": data_types},
        example={
            "data_json": {"text": "Well…   good luck…"},
            "raw_output": {'choices': [{'message': {'content': 'Well… good luck…', 'role': 'assistant',
                                                    'finish_reason': 'stop', 'index': 0, }}],
                           'created': 1683618639,
                           'id': 'chaxxpl-7EXXXXXwQNqPmXXXXXaYWkMtKNVjg',
                           'model': 'gpt-3.5-turbo-0301',
                           'object': 'chat.completion',
                           'usage': {'completion_tokens': 27,
                                     'prompt_tokens': 186,
                                     'total_tokens': 213}},
            "extra_fields": ["messages"],
            "data_audio_speak": ["data:audio/mpeg;base64,//tQxA/7n....VVVV.."],
            "data_url_speak": ["https://url..."],},
    )

    def __init__(self, api_text: Any,
                 api_audio: Any = None,
                 api_url: Any = None,
                 extra_fields: Any = None,
                 raw_output: Any = None,
                 **data: Any):
        super().__init__(**data)
        self.output = {
            "data_json": {"text": api_text},
            "data_audio_speak": api_audio,
            "data_url_speak": api_url,
            "extra_fields": extra_fields,
            "raw_output": raw_output
        }
      
    class Config:
        extra = Extra.allow

######################## Task: OCR ######################## TODO: Still in development

class OcrInput(BaseModel):
    """Input data for OCR task"""
    file_type: str = Field(..., description="File type image or pdf", example="image/png")
    file: str = Field(..., description="File in base64 format", example="data:image/png;base64,//tQxA-image-base64-format-string")

    class Config:
        extra = Extra.allow

class OcrOutput(BaseModel):
    """Output data for OCR task"""
    status: bool = True
    data_type: List[str] = Field(default=["data_image"], description="List of data types", enum=data_types)
    output: Dict[str, Any] = Field(
                                    default={},
                                    description="Output dictionary",
                                    keys={"enum": data_types},
                                    example={"data_image": ["data:image/png;base64,//tQxA-image-base64-format-string"]}
                                )

    def __init__(self, model_prediction: Any, data_type: List[str]=["data_image"], **data: Any):
        super().__init__(**data)
        self.data_type = data_type

        if "data_image" in self.data_type:
            self.output['data_image'] = model_prediction
        elif "data_pdf" in self.data_type:
            self.output['data_pdf'] = model_prediction
        else:
            raise ValueError("Invalid data_type, data_type must be data_image or data_pdf")
        
######################## Task: Translation ########################

class TranslationInput(BaseModel):
    """Input data for translation task"""
    text: List[str] = Field(..., description="List of texts", example=["This is an example text for translation."])
    source_language: str = Field(..., description="Source language", example="en")
    output_language: str = Field(..., description="Output language", example="de")

    class Config:
        extra = Extra.allow


class TranslationOutput(BaseModel):
    """Output data for translation task"""
    status: bool = True
    data_type: List[str] = Field(default=["data_json"], description="List of data types one or more item of data_types", enum=data_types)
    output: Dict[str, Any] = Field(
                                    default={"data_json": {"result": {'output_language':None, 'translated_text':None}}},
                                    description="Output dictionary",
                                    keys={"enum": data_types},
                                    example={
                                            "data_json": {
                                                            "result": {
                                                                        "output_language": "<output_language>",
                                                                        "translated_text": ["<translated_text>"]
                                                                    }}}
                                )

    def __init__(self, model_prediction: Any, output_language: str, **data: Any):
        super().__init__(**data)
        self.output['data_json']['result']['output_language'] = output_language
        self.output['data_json']['result']['translated_text'] = model_prediction
        
        class Config:
            extra = Extra.allow

######################## Task: Auto Speech Recognition ########################

class ASRInput(BaseModel):
    """Input data for ASR task"""
    data_type: List[str] = Field(default=["data_audio"], description="List of data types one or more item of data_types", enum=data_types)
    file: Optional[str] = Field(default=None, description="File in base64 format", example="data:audio/mp3;base64,//tQxA-audio-base64-format-audio-file...===")
    url: Optional[str] = Field(default=None, description="Audio file url", example="https://www.example.com/audio.mp3")
    translation: Optional[bool] = Field(default=False, description="Translation flag", example=True)
    language: Optional[str] = Field(default=None, description="Language code", example="en")

    class Config:
        extra = Extra.allow

class ASROutput(BaseModel):
    status: bool = True
    data_type: List[str] = Field(default=["data_json"], description="List of data types one or more item of data_types", enum=data_types)
    output: Dict[str, Any] = Field(default={"data_json": {"transcription":None, 
                                                          "translation":None, 
                                                          "segments":None, 
                                                          "language":None}}, description="Output dictionary", example={"data_json": {"transcription":"transcription_text", "translation": "translation_to_english_text_or_None", "segments": [{"text": "text_chunk_of_transcription.", "start": 0.0, "end": 3.0}], "language": "en"}})

    def __init__(self, transcription:list=None, translation=None, language:str="en", segments: Any=None, **data: Any):
        super().__init__(**data)
        self.output['data_json']['transcription'] = transcription
        self.output['data_json']['translation'] = translation
        self.output['data_json']['language'] = language
        self.output['data_json']['segments'] = segments

        class Config:
            extra = Extra.allow




# Author: AIMPED
# Date: 2023-March-12
# Description: This file contains the pipeline for de-identification of clinical notes
import pandas as pd
import random

def maskText(merged, text):
    """
    It masks the actual chunks in the text with their entity labels.
    parameters:
    ----------------
    merged: list of dict
    text: str
    return:
    ----------------
    masked_text: str
    """
    text_mask = text[:]
    for i in range(len(merged) - 1, -1, -1):
        text_mask = text_mask[:merged[i]['begin']] + f"<<{merged[i]['entity']}>>" + text_mask[merged[i]['end']:]
    return text_mask


def fakedChunk(fake_csv_path, merged ):
    """ 
    Randomly select entities from fake.csv file and add them to entities list align with true labels.
    parameters:
    ----------------
    fake_csv_path: str
    merged: list of dict
    return:
    ----------------
    merged: list of dict
    """


    fake_df = pd.read_csv(fake_csv_path, sep=',', encoding='utf8')
    for item in merged:
        item["faked_chunk"] = str(random.choice(fake_df[item['entity']]))
    return merged


def fakedText(merged, text):
    """
    Rewrite the text with faked chunks that comes from fakedChunk function.
    parameters:
    ----------------
    merged: list of dict
    text: str
    return:
    ----------------
    faked_text: str
    """
    
    faked_text = text[:]
    for i in range(len(merged) - 1, -1, -1):
        faked_text = faked_text[:merged[i]['begin']] + f"{merged[i]['faked_chunk']}" + faked_text[merged[i]['end']:]
    return faked_text


def deidentification(faked, masked, merged_results, text, fake_csv_path):
    """
    It masks the actual chunks in the text with their entity labels.
    parameters:
    ----------------
    faked: bool
    masked: bool
    merged_results: list of dict
    text: str
    fake_csv_path: str
    return:
    ----------------
    entities: list of dict
    masked_text: str
    faked_text: str
    """
    if faked and masked:
        text_mask = maskText(merged=merged_results, text=text)
        entities_with_faked_chunks = fakedChunk(fake_csv_path=fake_csv_path, merged=merged_results)
        faked_text = fakedText(merged=entities_with_faked_chunks, text=text)
        return {"entities": entities_with_faked_chunks, "masked_text":text_mask,"faked_text":faked_text}
    elif faked:
        entities_with_faked_chunks = fakedChunk(fake_csv_path=fake_csv_path,merged=merged_results)
        faked_text = fakedText(merged=entities_with_faked_chunks, text=text)
        return {"entities": entities_with_faked_chunks, "faked_text":faked_text}
    elif masked:
        text_mask = maskText(merged=merged_results, text=text)
        return {"entities": merged_results, "masked_text":text_mask}

    else:
        return {"entities": merged_results}
    

# visualizing the de-identified text

import random
import colorsys
from IPython.display import HTML, display

class DeidentificationVisualizer:
    def __init__(self):
        self.colors = [
            {"dark": "#5C5CE0", "light": "rgba(92, 92, 224, 0.12)"},
            {"dark": "#3DA74E", "light": "rgba(61, 167, 78, 0.12)"},
            {"dark": "#CE2783", "light": "rgba(206, 39, 131, 0.12)"},
            {"dark": "#D2B200", "light": "rgba(210, 178, 0, 0.12)"},
            {"dark": "#B130BD", "light": "rgba(177, 48, 189, 0.12)"},
            {"dark": "#16878C", "light": "rgba(22, 135, 140, 0.02)"},
            {"dark": "#7CC33F", "light": "rgba(124, 195, 63, 0.12)"},
            {"dark": "#864CCC", "light": "rgba(134, 76, 204, 0.12)"},
            {"dark": "#1473E6", "light": "rgba(20, 115, 230, 0.12)"},
            {"dark": "#D7373F", "light": "rgba(215, 55, 63, 0.12)"},
            {"dark": "#DA7B11", "light": "rgba(218, 123, 17, 0.12)"},
            {"dark": "#268E6C", "light": "rgba(38, 142, 108, 0.12)"},
        ]
        self.color_map = {}

    def generate_random_color(self):
        h, s, l = random.random(), 0.5 + random.random()/2.0, 0.4 + random.random()/5.0
        r, g, b = [int(256*i) for i in colorsys.hls_to_rgb(h, l, s)]
        dark = f"rgb({r}, {g}, {b})"
        light = f"rgba({r}, {g}, {b}, 0.12)"
        return {"dark": dark, "light": light}

    def get_color(self, entity):
        if entity not in self.color_map:
            if len(self.color_map) < len(self.colors):
                self.color_map[entity] = self.colors[len(self.color_map)]
            else:
                self.color_map[entity] = self.generate_random_color()
        return self.color_map[entity]

    def visualize_phi_entities(self, text, entities, is_short=False, show_size=None):
        html = ['<div style="padding: 14px;"><div dir="auto">']
        
        output = entities
        if is_short and show_size:
            output = [e for e in entities if e["end"] < show_size]
        
        last_index = 0
        for entity in output:
            plain_text = text[last_index:entity["begin"]]
            if plain_text:
                html.append(f'<span class="non-ner">{plain_text}</span>')
            
            color = self.get_color(entity["entity"])
            html.append(f'''
                <span class="entity-wrapper" style="background-color: {color['light']}; border-color: {color['dark']};">
                    <span class="entity-name">{entity["chunk"]}</span>
                    <span class="entity-type" style="background-color: {color['dark']};">{entity["entity"]}</span>
                </span>
            ''')
            last_index = entity["end"]
        
        if last_index < len(text):
            remaining_text = text[last_index:]
            if is_short and show_size and len(text) > show_size:
                remaining_text = remaining_text[:show_size - last_index] + "..."
            html.append(f'<span class="non-ner">{remaining_text}</span>')

        html.append('</div></div>')
        return ''.join(html)

    def visualize_anonymized(self, text, entities, is_short=False, show_size=None):
        html = ['<div style="padding: 14px;"><div dir="auto">']
        
        output = entities
        if is_short and show_size:
            output = [e for e in entities if e["end"] < show_size]
        
        last_index = 0
        for entity in output:
            plain_text = text[last_index:entity["begin"]]
            if plain_text:
                html.append(f'<span class="non-ner">{plain_text}</span>')
            
            color = self.get_color(entity["entity"])
            html.append(f'''
                <span class="entity-wrapper" style="background-color: {color['light']}; border-color: {color['dark']};">
                    <span class="entity-type" style="background-color: {color['dark']};">{entity["entity"]}</span>
                </span>
            ''')
            last_index = entity["end"]
        
        if last_index < len(text):
            remaining_text = text[last_index:]
            if is_short and show_size and len(text) > show_size:
                remaining_text = remaining_text[:show_size - last_index] + "..."
            html.append(f'<span class="non-ner">{remaining_text}</span>')

        html.append('</div></div>')
        return ''.join(html)

    def visualize_pseudonymized(self, text, entities, is_short=False, show_size=None):
        html = ['<div style="padding: 14px;"><div dir="auto">']
        
        output = entities
        if is_short and show_size:
            output = [e for e in entities if e["end"] < show_size]
        
        last_index = 0
        for entity in output:
            plain_text = text[last_index:entity["begin"]]
            if plain_text:
                html.append(f'<span class="non-ner">{plain_text}</span>')
            
            color = self.get_color(entity["entity"])
            html.append(f'''
                <span class="entity-wrapper" style="background-color: {color['light']}; border-color: {color['dark']};">
                    <span class="entity-name">{entity["faked_chunk"]}</span>
                    <span class="entity-type" style="background-color: {color['dark']};">{entity["entity"]}</span>
                </span>
            ''')
            last_index = entity["end"]
        
        if last_index < len(text):
            remaining_text = text[last_index:]
            if is_short and show_size and len(text) > show_size:
                remaining_text = remaining_text[:show_size - last_index] + "..."
            html.append(f'<span class="non-ner">{remaining_text}</span>')

        html.append('</div></div>')
        return ''.join(html)

    def visualize(self, text, entities, mode='phi_entities', is_short=False, show_size=None):
        css = '''
        <style>
        .non-ner {
            color: black;
            line-height: 2.2rem;
            font-size: 1rem;
        }
        .entity-wrapper {
            display: inline-flex;
            justify-content: space-between;
            text-align: center;
            border-radius: 5px;
            margin: 3px 2px;
            border: 1px solid;
            overflow: hidden;
            font-size: 1rem;
        }
        .entity-name {
            font-size: 0.9rem;
            line-height: 1.5rem;
            text-align: center;
            font-weight: 400;
            padding: 2px 5px;
            display: inline-block;
        }
        .entity-type {
            font-size: 0.9rem;
            line-height: 1.5rem;
            color: #fff;
            text-transform: uppercase;
            font-weight: 500;
            display: inline-block;
            padding: 2px 5px;
        }
        </style>
        '''
        
        if mode == 'phi_entities':
            html = self.visualize_phi_entities(text, entities, is_short, show_size)
        elif mode == 'anonymized':
            html = self.visualize_anonymized(text, entities, is_short, show_size)
        elif mode == 'pseudonymized':
            html = self.visualize_pseudonymized(text, entities, is_short, show_size)
        else:
            raise ValueError("Invalid mode. Choose 'phi_entities', 'anonymized', or 'pseudonymized'.")
        
        return css + html

    def display_visualization(self, text, entities, mode='phi_entities', is_short=False, show_size=None):
        html_output = self.visualize(text, entities, mode, is_short, show_size)
        display(HTML(html_output))

## Example usage
# if __name__ == "__main__":
#     visualizer = DeidentificationVisualizer()
    
#     sample_text = "John Smith was admitted on 2023-05-15 with severe abdominal pain."
#     sample_entities = [
#         {"begin": 0, "end": 10, "chunk": "John Smith", "entity": "PERSON", "faked_chunk": "Alice Johnson"},
#         {"begin": 27, "end": 37, "chunk": "2023-05-15", "entity": "DATE", "faked_chunk": "2023-06-20"}
#     ]
    
#     print("PHI Entities Visualization:")
#     visualizer.display_visualization(sample_text, sample_entities, mode='phi_entities')
    
#     print("\nAnonymized Visualization:")
#     visualizer.display_visualization(sample_text, sample_entities, mode='anonymized')
    
#     print("\nPseudonymized Visualization:")
#     visualizer.display_visualization(sample_text, sample_entities, mode='pseudonymized')


# Author: AIMPED
# Date: 2023-March-11
# Description: NER model results

import torch
import pandas as pd
import numpy as np
import os
import re
import json
from aimped.nlp.tokenizer import sentence_tokenizer, word_tokenizer

def NerModelResults(sents_tokens_list, sentences, tokenizer, model, text, device,
                    assertion_relation=False):
    """
    It returns the NER model results of a text.
    Parameters
    ----------
    sents_tokens_list : list
    tokenizer : transformers.PreTrainedTokenizer
    model : transformers.PreTrainedModel
    text : str
    device : torch.device
    assertion_relation : bool, optional
        The default is False.
    sentences : list, optional
        The default is []. Only used if assertion_relation is True

    Returns
    -------
    tokens : list
    preds : list
    probs : list
    begins : list
    ends : list
    sent_begins : list
        Only returned if assertion_relation is True
    sent_ends : list
        Only returned if assertion_relation is True
    sent_idxs : list
        Only returned if assertion_relation is True
    """

    start = 0
    tokens, probs, begins, ends, preds, sent_begins, sent_ends, sent_idxs = [], [], [], [], [], [], [], []

    for sentence_idx, sent_token_list in enumerate(sents_tokens_list):
        start_sent = 0
        start = text.find(sentences[sentence_idx], start)
        model_inputs = tokenizer(sent_token_list, is_split_into_words=True, truncation=True,
                                 padding=False, max_length=512, return_tensors="pt").to(device)
        word_ids = model_inputs.word_ids()  # sub tokenlar sent_token_list deki hangi idxteki tokena ait
        # ornek word_ids = [None, 0, 1, 2, 2, 2, 3, 4, 5, 6, 7, 8, 9, 10, 10, 10, 11, 12, 13, 14, 15, None]
        outputs = model(**model_inputs)
        probabilities = torch.nn.functional.softmax(outputs.logits, dim=-1).tolist()[0]
        predictions = outputs.logits.argmax(dim=-1).tolist()[0]
        idx = 1
        while idx < len(word_ids) - 1:  # sondaki None icin islem yapmamak icin -1 yapildi
            word_id1 = word_ids[idx]
            word_id2 = word_ids[idx + 1]
            label = model.config.id2label[predictions[idx]]
            prob = max(probabilities[idx])
            if word_id1 == word_id2:
                while word_id1 == word_ids[idx]:
                    idx += 1
                idx -= 1

            token = sent_token_list[word_ids[idx]]
            begin = text.find(token, start)
            end = begin + len(token)
            tokens.append(token)
            begins.append(begin)
            ends.append(end)
            preds.append(label)
            probs.append(prob)
            start = end
            if assertion_relation:
                sentence_begin = sentences[sentence_idx].find(token, start_sent)
                sentence_end = sentence_begin + len(token)
                sent_begins.append(sentence_begin)
                sent_ends.append(sentence_end)
                sent_idxs.append(sentence_idx)
                start_sent = sentence_end
            idx += 1
    if not assertion_relation:
        return tokens, preds, probs, begins, ends
    else:
        return tokens, preds, probs, begins, ends, sent_begins, sent_ends, sent_idxs


import random
import colorsys
from IPython.display import HTML, display

class NERVisualizer:
    def __init__(self):
        self.colors = [
            {"dark": "#5C5CE0", "light": "rgba(92, 92, 224, 0.12)"},
            {"dark": "#3DA74E", "light": "rgba(61, 167, 78, 0.12)"},
            {"dark": "#CE2783", "light": "rgba(206, 39, 131, 0.12)"},
            {"dark": "#D2B200", "light": "rgba(210, 178, 0, 0.12)"},
            {"dark": "#B130BD", "light": "rgba(177, 48, 189, 0.12)"},
            {"dark": "#16878C", "light": "rgba(22, 135, 140, 0.02)"},
            {"dark": "#7CC33F", "light": "rgba(124, 195, 63, 0.12)"},
            {"dark": "#864CCC", "light": "rgba(134, 76, 204, 0.12)"},
            {"dark": "#1473E6", "light": "rgba(20, 115, 230, 0.12)"},
            {"dark": "#D7373F", "light": "rgba(215, 55, 63, 0.12)"},
            {"dark": "#DA7B11", "light": "rgba(218, 123, 17, 0.12)"},
            {"dark": "#268E6C", "light": "rgba(38, 142, 108, 0.12)"},
        ]
        self.color_map = {}

    def generate_random_color(self):
        h, s, l = random.random(), 0.5 + random.random()/2.0, 0.4 + random.random()/5.0
        r, g, b = [int(256*i) for i in colorsys.hls_to_rgb(h, l, s)]
        dark = f"rgb({r}, {g}, {b})"
        light = f"rgba({r}, {g}, {b}, 0.12)"
        return {"dark": dark, "light": light}

    def get_color(self, entity):
        if entity not in self.color_map:
            if len(self.color_map) < len(self.colors):
                self.color_map[entity] = self.colors[len(self.color_map)]
            else:
                self.color_map[entity] = self.generate_random_color()
        return self.color_map[entity]

    def visualize_entities(self, text, entities, is_short=False, show_size=None):
        html = ['<div style="padding: 14px;"><div dir="auto">']
        
        output = entities
        if is_short and show_size:
            output = [e for e in entities if e["end"] < show_size]
        
        last_index = 0
        for entity in output:
            plain_text = text[last_index:entity["begin"]]
            if plain_text:
                html.append(f'<span class="non-ner">{plain_text}</span>')
            
            color = self.get_color(entity["entity"])
            html.append(f'''
                <span class="entity-wrapper" style="background-color: {color['light']}; border-color: {color['dark']};">
                    <span class="entity-name">{entity["chunk"]}</span>
                    <span class="entity-type" style="background-color: {color['dark']};">{entity["entity"]}</span>
                </span>
            ''')
            last_index = entity["end"]
        
        if last_index < len(text):
            remaining_text = text[last_index:]
            if is_short and show_size and len(text) > show_size:
                remaining_text = remaining_text[:show_size - last_index] + "..."
            html.append(f'<span class="non-ner">{remaining_text}</span>')

        html.append('</div></div>')
        return ''.join(html)





    def visualize(self, text, entities, is_short=False, show_size=None):
        css = '''
        <style>
        .non-ner {
            color: black;
            line-height: 2.2rem;
            font-size: 1rem;
        }
        .entity-wrapper {
            display: inline-flex;
            justify-content: space-between;
            text-align: center;
            border-radius: 5px;
            margin: 3px 2px;
            border: 1px solid;
            overflow: hidden;
            font-size: 1rem;
        }
        .entity-name {
            font-size: 0.9rem;
            line-height: 1.5rem;
            text-align: center;
            font-weight: 400;
            padding: 2px 5px;
            display: inline-block;
        }
        .entity-type {
            font-size: 0.9rem;
            line-height: 1.5rem;
            color: #fff;
            text-transform: uppercase;
            font-weight: 500;
            display: inline-block;
            padding: 2px 5px;
        }
        </style>
        '''
        
        html = self.visualize_entities(text, entities, is_short, show_size)

        return css + html

    def display_visualization(self, text, entities, is_short=False, show_size=None):
        html_output = self.visualize(text, entities, is_short, show_size)
        display(HTML(html_output))




# Author: Russell C.
# Date: 2023-Sep-17
# Description: This is the utils file for the sound module.

import os
import io
import logging

try:
    from pydub import AudioSegment
except ImportError:
    raise ImportError("The 'pydub' library is required to use this module. You can install it with 'pip install pydub'.")

# Default values
DEFAULT_SAMPLE_RATE = 16000
DEFAULT_CHANNELS = 1

def mp3_mp4_to_wav(input_file, output_file, sample_rate=DEFAULT_SAMPLE_RATE):
    """
    Converts MP3/MP4 files to WAV format with optional sample rate conversion and mono channel.
    
    Args:
        input_file (str): Input audio file path (MP3/MP4).
        output_file (str): Output WAV file path.
        sample_rate (int): Target sample rate (default: 16000).
    """
    try:
        audio = AudioSegment.from_file(input_file)
        audio = audio.set_channels(DEFAULT_CHANNELS)  # Convert to mono channel
        audio = audio.set_frame_rate(sample_rate)
        audio.export(output_file, format="wav")
        logging.info(f"Converted {input_file} to {output_file} with sample rate {sample_rate} and mono channel")
    except Exception as e:
        logging.error(f"Error converting {input_file} to WAV: {e}")

def cut_audio(input_file, output_file, start_time_ms, end_time_ms):
    """
    Cuts a portion of an audio file and saves it as a new WAV file.
    
    Args:
        input_file (str): Input audio file path (WAV).
        output_file (str): Output WAV file path.
        start_time_ms (int): Start time of the cut in milliseconds.
        end_time_ms (int): End time of the cut in milliseconds.
    """
    try:
        audio = AudioSegment.from_file(input_file)
        cut_audio = audio[start_time_ms:end_time_ms]
        cut_audio.export(output_file, format="wav")
        logging.info(f"Cut {input_file} and saved as {output_file}")
    except Exception as e:
        logging.error(f"Error cutting {input_file}: {e}")

def change_sample_rate(input_file, output_file, new_sample_rate=DEFAULT_SAMPLE_RATE):
    """
    Changes the sample rate of an audio file and saves it as a new WAV file.
    
    Args:
        input_file (str): Input audio file path (WAV).
        output_file (str): Output WAV file path.
        new_sample_rate (int): New sample rate (default: 16000).
    """
    try:
        audio = AudioSegment.from_file(input_file)
        audio = audio.set_frame_rate(new_sample_rate)
        audio.export(output_file, format="wav")
        logging.info(f"Changed sample rate of {input_file} to {new_sample_rate} and saved as {output_file}")
    except Exception as e:
        logging.error(f"Error changing sample rate of {input_file}: {e}")

def change_channels(input_file, output_file, num_channels=DEFAULT_CHANNELS):
    """
    Changes the number of channels of an audio file and saves it as a new WAV file.
    
    Args:
        input_file (str): Input audio file path (WAV).
        output_file (str): Output WAV file path.
        num_channels (int): Number of channels (default: 1, mono).
    """
    try:
        audio = AudioSegment.from_file(input_file)
        audio = audio.set_channels(num_channels)
        audio.export(output_file, format="wav")
        logging.info(f"Changed channels of {input_file} to {num_channels} channel(s) and saved as {output_file}")
    except Exception as e:
        logging.error(f"Error changing channels of {input_file}: {e}")

def get_audio_duration(input_file):
    """
    Gets the duration of an audio file in seconds.

    Args:
        input_file (str or bytes): Input audio file path (MP3/MP4/WAV) or binary data.

    Returns:
        float: Duration in seconds, or None if an error occurs.
    """
    if isinstance(input_file, str):
        # If input_file is a string (file path)
        try:
            audio = AudioSegment.from_file(input_file)
            duration_ms = len(audio)
            duration_seconds = duration_ms / 1000.0  # Convert to seconds
            return duration_seconds
        except Exception as e:
            error_message = f"Error getting duration of {input_file}: {e}"
            logging.error(error_message)
            return None
    elif isinstance(input_file, bytes):
        # If input_file is binary data
        try:
            audio = AudioSegment.from_file(io.BytesIO(input_file))
            duration_ms = len(audio)
            duration_seconds = duration_ms / 1000.0  # Convert to seconds
            return duration_seconds
        except Exception as e:
            error_message = f"Error getting duration from binary data: {e}"
            logging.error(error_message)
            return None
    else:
        error_message = "Unsupported input type. Please provide a file path (str) or binary data (bytes)."
        logging.error(error_message)
        return None

def get_audio_info(input_file):
    """
    Gets information about an audio file using pydub and Python standard libraries.
    
    Args:
        input_file (str): Input audio file path (MP3/MP4/WAV).
    
    Returns:
        dict: A dictionary containing audio information.
            - 'duration' (float): Duration in seconds.
            - 'sample_rate' (int): Sample rate in Hz.
            - 'channels' (int): Number of channels (1 for mono, 2 for stereo).
            - 'format' (str): Audio format (e.g., 'mp3', 'wav').
            - 'size' (int): File size in bytes.
    """
    audio_info = {}
    
    if not os.path.exists(input_file):
        error_message = f"File {input_file} does not exist."
        logging.error(error_message)
        return None
    
    try:
        audio = AudioSegment.from_file(input_file)
        
        # Duration in seconds
        duration = len(audio) / 1000  # Convert to seconds
        audio_info['duration'] = duration
        
        # Sample rate
        audio_info['sample_rate'] = audio.frame_rate
        
        # Number of channels
        audio_info['channels'] = audio.channels
        
        # Audio format
        audio_info['format'] = os.path.splitext(input_file)[-1][1:].lower()
        
        # File size
        audio_info['size'] = os.path.getsize(input_file)
        
        return audio_info
    except Exception as e:
        error_message = f"Error getting information for {input_file}: {e}"
        logging.error(error_message)
        return None

# if __name__ == "__main__":
#     input_file = "input.mp3"
#     output_file = "output.wav"

#     audio_info = get_audio_info(input_file)
#     if audio_info is not None:
#         logging.info(f"Audio info for {input_file}: {audio_info}")

#     mp3_mp4_to_wav(input_file, output_file)  # Using default sample rate and mono channel
#     cut_audio(output_file, "cut_output.wav", 1000, 4000)
#     change_sample_rate(output_file, "sample_rate_output.wav")  # Using default sample rate
#     change_channels(output_file, "channels_output.wav")  # Using default mono channel

#     duration = get_audio_duration(output_file)
#     if duration is not None:
#         logging.info(f"Duration of {output_file}: {duration} seconds")

#     audio_info = get_audio_info(output_file)
#     if audio_info is not None:
#         logging.info(f"Audio info for {output_file}: {audio_info}")



import requests
from datetime import datetime


class Connect:
    """Class representing a connection to A3m API."""

    default_option = {
        'base_url': 'base_url',
        'scope': 'scope',
    }

    default_token_data = {
        'access_token': '',
        'refresh_token': '',
        'token_type': '',
        'expires_in': 0,
        'scope': '',
    }

    def __init__(self, user_key: str, user_secret: str, options):
        """
        Initialize A3mConnect instance.

        Parameters:
        - user_key (str): User key for authentication.
        - user_secret (str): User secret for authentication.
        - options (dict): Additional options for the connection.
        """
        self.user_key = user_key
        self.user_secret = user_secret
        # Merge default options with provided options
        self.options = {**self.default_option, **options}
        self.is_init = False
        self.token_data = {**self.default_token_data}
        self.time_data = {
            'token_time': None,
            'ref_token_time': None,
        }

    def get_user_keys(self):
        """
        Get user keys.

        Returns:
        dict: Dictionary containing user_key and user_secret.
        """
        return {
            'user_key': self.user_key,
            'user_secret': self.user_secret,
        }

    def get_options(self):
        """
        Get connection options.

        Returns:
        dict: Connection options.
        """
        return self.options

    def get_token_data(self):
        """
        Get token data.

        Returns:
        dict: Token data.
        """
        if not self.is_init:
            self.init()
        return self.token_data

    def is_token_expired(self):
        """
        Check if the access token is expired.

        Returns:
        bool: True if the access token is expired, False otherwise.
        """
        cur_date = datetime.utcnow()
        return (
            (cur_date - self.time_data['token_time']).total_seconds() >
            (self.token_data['expires_in'] - 180)
        )

    def is_ref_token_expired(self):
        """
        Check if the refresh token is expired.

        Returns:
        bool: True if the refresh token is expired, False otherwise.
        """
        cur_date = datetime.utcnow()
        return (cur_date - self.time_data['ref_token_time']).days > 6

    def get_all_data(self):
        """
        Get all data related to the connection.

        Returns:
        A3mConnect: A3mConnect instance.
        """
        if not self.is_init:
            self._init()
        return self

    def is_connected(self):
        """
        Check if the instance is connected and handle token refreshing.

        Returns:
        A3mConnect: A3mConnect instance.
        """
        try:
            if not self.is_init:
                return self.init()
            elif self.is_ref_token_expired():
                return self.init()
            elif self.is_token_expired():
                return self.refresh_access_token()
            else:
                return self
        except Exception as e:
            raise Exception(e)

    def init(self):
        """
        Initialize the connection and retrieve access and refresh tokens.

        Returns:
        A3mConnect: A3mConnect instance.
        """
        try:
            params = {
                'grant_type': 'password',
                'username': self.user_key,
                'password': self.user_secret,
            }
            headers = {
                'Content-Type': 'application/x-www-form-urlencoded',
            }

            response = requests.post(
                f"{self.options['base_url']}/token", headers=headers,
                data=params
            )

            response_data = response.json()
            self.token_data = {**self.token_data, **response_data}
            self.time_data = {
                'token_time': datetime.utcnow(),
                'ref_token_time': datetime.utcnow(),
            }

            self.is_init = True
            return self
        except requests.exceptions.RequestException as e:
            raise Exception(f"Error during initialization: {e}")

    def refresh_access_token(self):
        """
        Refresh the access token using the refresh token.

        Returns:
        A3mConnect: A3mConnect instance.
        """
        try:
            params = {
                'grant_type': 'refresh_token',
                'refresh_token': self.token_data['refresh_token'],
            }

            headers = {
                'Content-Type': 'application/x-www-form-urlencoded',
            }

            response = requests.post(
                f"{self.options['base_url']}/token", headers=headers,
                data=params
            )

            response_data = response.json()
            self.token_data = {**self.token_data, **response_data}
            self.time_data['token_time'] = datetime.utcnow()
            return self
        except requests.exceptions.RequestException as e:
            raise Exception(f"Error refreshing access token: {e}")



# medical_coding.py

import random
import colorsys
from IPython.display import HTML, display

class MedicalCodingVisualizer:
    def __init__(self):
        self.colors = [
            {"dark": "#5C5CE0", "light": "rgba(92, 92, 224, 0.12)"},
            {"dark": "#3DA74E", "light": "rgba(61, 167, 78, 0.12)"},
            {"dark": "#CE2783", "light": "rgba(206, 39, 131, 0.12)"},
            {"dark": "#D2B200", "light": "rgba(210, 178, 0, 0.12)"},
            {"dark": "#B130BD", "light": "rgba(177, 48, 189, 0.12)"},
            {"dark": "#16878C", "light": "rgba(22, 135, 140, 0.02)"},
            {"dark": "#7CC33F", "light": "rgba(124, 195, 63, 0.12)"},
            {"dark": "#864CCC", "light": "rgba(134, 76, 204, 0.12)"},
            {"dark": "#1473E6", "light": "rgba(20, 115, 230, 0.12)"},
            {"dark": "#D7373F", "light": "rgba(215, 55, 63, 0.12)"},
            {"dark": "#DA7B11", "light": "rgba(218, 123, 17, 0.12)"},
            {"dark": "#268E6C", "light": "rgba(38, 142, 108, 0.12)"},
        ]
        self.color_map = {}
        self.assertion_color_map = {}

    def generate_random_color(self):
        h, s, l = random.random(), 0.5 + random.random()/2.0, 0.4 + random.random()/5.0
        r, g, b = [int(256*i) for i in colorsys.hls_to_rgb(h, l, s)]
        dark = f"rgb({r}, {g}, {b})"
        light = f"rgba({r}, {g}, {b}, 0.12)"
        return {"dark": dark, "light": light}

    def get_color(self, entity):
        if entity not in self.color_map:
            if len(self.color_map) < len(self.colors):
                self.color_map[entity] = self.colors[len(self.color_map)]
            else:
                self.color_map[entity] = self.generate_random_color()
        return self.color_map[entity]

    def get_assertion_color(self, assertion):
        if assertion not in self.assertion_color_map:
            if len(self.assertion_color_map) < len(self.colors):
                self.assertion_color_map[assertion] = self.colors[-(len(self.assertion_color_map) + 1)]["dark"]
            else:
                self.assertion_color_map[assertion] = self.generate_random_color()["dark"]
        return self.assertion_color_map[assertion]

    def visualize(self, text, data, is_short=False, show_size=None):
        output = []
        html = ['<div style="padding: 14px;"><div dir="auto">']
        
        if is_short and show_size:
            for entity in data:
                if entity["end"] < show_size:
                    output.append(entity)
                else:
                    break
        else:
            output = data

        last_index = 0
        for entity in output:
            plain_text = text[last_index:entity["begin"]]
            if plain_text:
                html.append(f'<span class="non-ner">{plain_text}</span>')

            color = self.get_color(entity["entity"])
            assertion_color = self.get_assertion_color(entity["description"])
            
            html.append(f'''
                <span class="entity-wrapper-outer" style="border-color: {color['dark']};">
                    <span class="entity-wrapper" style="background-color: {color['light']};">
                        <span class="entity-type" style="background-color: {color['dark']};">{entity["code"]}</span>
                        <span class="entity-name">{entity["chunk"]}</span>
                        <span class="entity-type" style="background-color: {color['dark']};">{entity["entity"]}</span>
                    </span>
                    <span class="entity-type-assertion" style="background-color: {assertion_color};">{entity["description"]}</span>
                </span>
            ''')

            last_index = entity["end"]

        if last_index < len(text):
            remaining_text = text[last_index:]
            if is_short and show_size and len(text) > show_size:
                remaining_text = remaining_text[:show_size - last_index] + "..."
            html.append(f'<span class="non-ner">{remaining_text}</span>')

        html.append('</div></div>')

        css = '''
        <style>
        .non-ner {
            color: black;
            line-height: 3.4rem;
            font-size: 1rem;
        }
        .entity-wrapper-outer {
            display: inline-grid;
            text-align: center;
            border-radius: 4px;
            margin: 5px 2px;
            border: 1px solid;
            font-size: 1rem;
        }
        .entity-wrapper {
            display: inline-flex;
            text-align: center;
            justify-content: space-between;
            font-size: 1rem;
        }
        .entity-name {
            font-size: 0.9rem;
            line-height: 1.5rem;
            text-align: center;
            font-weight: 400;
            padding: 2px 5px;
            display: inline-block;
        }
        .entity-type, .entity-type-assertion {
            font-size: 0.9rem;
            line-height: 1.5rem;
            color: #fff;
            text-transform: uppercase;
            font-weight: 500;
            display: inline-block;
            padding: 3px 5px;
        }
        </style>
        '''

        return css + ''.join(html)

    def display_visualization(self, text, data, is_short=False, show_size=None):
        html_output = self.visualize(text, data, is_short, show_size)
        display(HTML(html_output))


import requests
from typing import Any, Dict, Callable, Optional
from datetime import datetime
import os
import time
from aimped.services.auth_service import Connect
# from aimped.a3mconnect import A3mConnect


class AimpedAPI(Connect):
    """
    Class representing a connection to the Aimped API, extending A3mConnect.

    Args:
        user_key (str): User key for authentication.
        user_secret (str): User secret for authentication.
        options (dict): Additional options for the connection.
    """

    def __init__(self, user_key, user_secret, options):
        """
        Initialize Aimped instance.

        Args:
            user_key (str): User key for authentication.
            user_secret (str): User secret for authentication.
            options (dict): Additional options for the connection.
        """
        super().__init__(user_key, user_secret, options)

    def run_model(self, model_id: int, payload: Dict[str, Any]) -> Any:
        """
        Run a model prediction.

        Args:
            model_id (int): The ID of the model to run.
            payload (dict): Input payload for the model.

        Returns:
            Any: Result of the model prediction.
        """
        try:
            if self.is_connected():
                headers = {
                    'Authorization': f'Bearer {self.token_data["access_token"]}',
                    'Content-Type': 'application/json',
                }
                # print("istek atıldı")
                result = requests.request(
                    'POST',
                    f'{self.options["base_url"]}/pub/backend/api/v1/model_run_prediction/{model_id}/',
                    json=payload, headers=headers
                )

                return result.json()

        except requests.exceptions.RequestException as e:
            raise Exception(e)

    def get_pod_log_result(self, model_id: int) -> Dict[str, Any]:
        """
        Get the result of the pod logs for a specific model.

        Args:
            model_id (int): The ID of the model.

        Returns:
            dict: Result of the pod logs.
        """
        try:
            result = requests.request(
                'GET',
                f'{self.options["base_url"]}/pub/backend/get_pod_log?model_id={model_id}&instance_id=&is_dedicated='
            )
            return result.json()
        except requests.exceptions.RequestException as e:
            raise Exception(e)

    def run_model_callback(
        self,
        model_id: int,
        payload: Dict[str, Any],
        callback: Callable[[str, str, str, Optional[str]], str]
    ) -> Any:
        """
        Run a model with a callback function to monitor its progress.

        Args:
            model_id (int): The ID of the model to run.
            payload (dict): Input payload for the model.
            callback (Callable): Callback function to monitor model progress.

        Returns:
            Any: Result of the model prediction.
        """
        try:
            if self.is_connected():
                callback(
                    event='start',
                    message='start model run',
                    time=datetime.now().strftime("%Y-%m-%d %H:%M:%S")
                )

                result = self.run_model(model_id, payload)
                return result

                keep_running = True
                while keep_running:
                    time.sleep(15)
                    pod_log_result = self.get_pod_log_result(model_id)

                    waiting = pod_log_result.get('waiting', '').replace(
                        'Container', '') if pod_log_result.get('waiting') else None

                    if waiting:
                        if waiting == 'CrashLoopBackOff':
                            callback(
                                event='error',
                                message='Model is CrashLoopBackOff.',
                                time=datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                            )
                            keep_running = False
                        else:
                            callback(
                                event='proccess',
                                message=f'Model is {waiting}',
                                time=datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                            )
                    elif pod_log_result.get('error'):
                        callback(
                            event='error',
                            message=pod_log_result['error'],
                            time=datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                        )
                        keep_running = False

                    elif pod_log_result.get('running'):
                        time.sleep(10)
                        result = self.run_model(model_id, payload)
                        callback(
                            event='end',
                            message='ok',
                            data=result,
                            time=datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                        )
                        keep_running = False
                    else:
                        callback(
                            event='proccess',
                            message='Waiting for model to be ready',
                            time=datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                        )

        except Exception as e:
            callback(
                event='error',
                message=str(e),
                time=datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            )

    def file_upload(self, model_id: int, source: str):
        """
        Upload a file to the specified model.

        Args:
            model_id (int): The ID of the model.
            filepath (str): Path to the file to be uploaded.

        Returns:
            dict: Result of the file upload operation.
        """
        try:
            if self.is_connected():
                with open(source, 'rb') as file:
                    files = {'file': (os.path.basename(source), file)}
                    headers = {
                        'Authorization': f'Bearer {self.token_data["access_token"]}'
                    }
                    result = requests.post(
                        f'{self.options["base_url"]}/pub/backend/api/v1/file_upload/{model_id}',
                        files=files,
                        headers=headers,
                    )
                    return result.json()
        except requests.exceptions.RequestException as e:
            raise Exception(e)

    def file_download_and_save(self, source: str, target: str):
        """
        Download a file from the specified source and save it to the target.

        Args:
            filepath (str): Path to the file to be downloaded.
            destination (str): Path where the downloaded file will be saved.

        Returns:
            str: Path of the downloaded file.
        """
        try:
            if self.is_connected():
                headers = {
                    'Authorization': f'Bearer {self.token_data["access_token"]}',
                }
                result = requests.get(
                    f'{self.options["base_url"]}/pub/backend/api/v1/file_download?file={source}',
                    headers=headers,
                    stream=True,
                )

                with open(target, 'wb') as file:
                    for chunk in result.iter_content(chunk_size=128):
                        file.write(chunk)
                return target
        except requests.exceptions.RequestException as e:
            raise Exception(e)



# Author AIMPED
# Date 2023-March-11
# Description: This file contains the pipeline wrapper for NER, Assertion and De-identification models.

from aimped.nlp.ner import NerModelResults
from aimped.nlp.deid import maskText, fakedChunk, fakedText, deidentification
from aimped.nlp.assertion import AssertionAnnotateSentence, AssertionModelResults
from aimped.nlp.chunker import ChunkMerger
from aimped.nlp.regex_parser import RegexNerParser, RegexModelNerMerger, RegexModelOutputMerger
from aimped.nlp.relation import RelationResults, RelationAnnotateSentence
import glob
class Pipeline:
    """
    It returns the ner results of a text.
    parameters:
    ----------------
    task: str
    text: str
    model: transformers.modeling_utils.PreTrainedModel
    tokenizer: transformers.tokenization_utils_base.PreTrainedTokenizer
    *args: arguments
    **kwargs: keyword arguments
    return:
    ----------------
    results: list of dict
    """

    def __init__(self, tokenizer, model, device='cpu'):
        """Initialize the pipeline class."""
        self.tokenizer = tokenizer
        self.model = model
        self.device = device

    def ner_result(self, text, sents_tokens_list, sentences, assertion_relation=False):
        """It returns the ner results of a text.
        parameters:
        ----------------
        text: str
        sents_tokens_list: list of list of str
        assertion_relation: bool
        sentences: list of str
        return:
        ----------------
        ner_results: list of dict"""
        ner_results = NerModelResults(text=text,
                                      tokenizer=self.tokenizer,
                                      model=self.model,
                                      device=self.device,
                                      sents_tokens_list=sents_tokens_list,
                                      sentences=sentences,
                                      assertion_relation=assertion_relation
                                      )

        return ner_results

    def deid_result(self, text, merged_results, fake_csv_path, faked=False, masked=False):
        """It returns the deid results of a text.
        parameters:
        ----------------
        text: str
        merged_results: list of dict
        fake_csv_path: str
        faked: bool
        masked: bool
        return:
        ----------------
        results: list of dict
        """
        results = deidentification(text=text,
                                   merged_results=merged_results,
                                   fake_csv_path=fake_csv_path,
                                   faked=faked,
                                   masked=masked)
        return results

    def assertion_result(self, ner_results, sentences, classifier, assertion_white_label_list,resolver=False):
        """It returns the assertion results of a text.
        parameters:
        ----------------
        text: str
        ner_results: list of dict
        sentences: list of str
        classifier: str
        return:
        ----------------
        results: list of dict
        """
        results = AssertionModelResults(ner_results=ner_results,
                                        sentences=sentences,
                                        classifier=classifier,
                                        assertion_white_label_list=assertion_white_label_list,
                                        resolver=resolver
                                        )
        return results

    def chunker_result(self, text, white_label_list, tokens, preds, probs, begins, ends,
                       assertion_relation=False, sent_begins=[], sent_ends=[], sent_idxs=[]):
        """It returns the merged chunks of a text.
        parameters:
        ----------------
        text: str
        white_label_list: list of str
        tokens: list of str
        preds: list of str
        probs: list of float
        begins: list of int
        ends: list of int
        assertion_relation: bool
        sent_begins: list of int
        sent_ends: list of int
        sent_idxs: list of int
        return:
        ----------------
        results: list of dict
        """
        results = ChunkMerger(text=text,
                              white_label_list=white_label_list,
                              tokens=tokens,
                              preds=preds,
                              probs=probs,
                              begins=begins,
                              ends=ends,
                              assertion_relation=assertion_relation,
                              sent_begins=sent_begins,
                              sent_ends=sent_ends,
                              sent_idxs=sent_idxs)
        return results

    def regex_model_output_merger(self, regex_json_files_path, model_results, text, white_label_list):
        """It returns the regex results of a text.
        parameters:
        ----------------
        regex_json_files_path: str
        model_results: list of dict
        text: str
        white_label_list: list of str
        return:
        ----------------
        results: list of dict
        """

        regex_json_files_path_list = glob.glob(f"{regex_json_files_path}/*.json")

        merged_results = RegexModelOutputMerger(regex_json_files_path_list=regex_json_files_path_list,
                                                model_results=model_results,
                                                text=text,
                                                white_label_list=white_label_list)
        return merged_results

    def relation_result(self, sentences, ner_chunk_results, relation_classifier,
                        relation_white_label_list, relation_pairs, return_svg=False):
        """It returns the relation results of a text.
        parameters:
        ----------------
        sentences: list of str
        ner_chunk_results: list of dict
        relation_classifier: str
        ner_white_label_list: list of str
        relation_white_label_list: list of str
        one_to_many: bool = True
        one_label: str = None
        return_svg: bool = False
        return:
        ----------------
        results: list of dict
        """
        results = RelationResults(sentences=sentences,
                                  ner_chunk_results=ner_chunk_results,
                                  relation_classifier=relation_classifier,
                                  relation_white_label_list=relation_white_label_list,
                                  relation_pairs =relation_pairs,
                                  return_svg=return_svg)
        return results

 
   

    def __str__(self) -> str:
        """Return the string representation of the pipeline."""
        return f"Pipeline(model={self.model}, tokenizer={self.tokenizer})"






# -*- coding: utf-8 -*-
"""ner_cls_report

Automatically generated by Colaboratory.

Original file is located at
    https://colab.research.google.com/drive/1LJIu53E46CcoGLqcAhgav970B6X-coYm
"""

import pandas as pd
import warnings
warnings.filterwarnings('ignore')

try:        
    from seqeval.metrics import classification_report
except:
    print('seqeval is not installed. Please install it with pip install seqeval')

from sklearn.metrics import classification_report
    
def ReadConll(filename):
    df = pd.read_csv(filename,
                    sep = ' ', header = None, keep_default_na = False,
                    names = ['words', 'pos', 'chunk', 'labels'],
                    quoting = 3, 
                     skip_blank_lines = False, 
                     encoding="utf8")
    df = df[~df['words'].astype(str).str.startswith('-DOCSTART-')] # Remove the -DOCSTART- header
    df['sentence_id'] = (df.words == '').cumsum()
    return df[df.words != '']

def ClsReportNerModel(test_conll_path, tokenizer, model, device):


    test = ReadConll(test_conll_path)
    sents_tokens_list, truth_list = [],[]
    model = model.to(device)
    for i in test.sentence_id.unique():
        sents_tokens_list.append(list(test[test.sentence_id == i].words))
        truth_list.append(list(test[test.sentence_id == i].labels))
    tokens,preds,truths= [],[],[]
    for sentence_idx, sent_token_list in enumerate(sents_tokens_list):
        model_inputs = tokenizer(sent_token_list, is_split_into_words = True, truncation=True,
                                        padding=False, max_length=512, return_tensors="pt").to(device)
        word_ids = model_inputs.word_ids() # sub tokenlar sent_token_list deki hangi idxteki tokena ait
        # ornek word_ids = [None, 0, 1, 2, 2, 2, 3, 4, 5, 6, 7, 8, 9, 10, 10, 10, 11, 12, 13, 14, 15, None]
        outputs = model(**model_inputs)
        predictions = outputs.logits.argmax(dim=-1).tolist()[0]
        idx = 1
        while idx < len(word_ids)-1: # sondaki None icin islem yapmamak icin -1 yapildi
            word_id1 = word_ids[idx]
            word_id2 = word_ids[idx + 1]
            label = model.config.id2label[predictions[idx]]
            if word_id1 == word_id2:
                while word_id1 == word_ids[idx]:                
                    idx +=1
                idx -=1

            token = sent_token_list[word_ids[idx]]
            truth = truth_list[sentence_idx][word_ids[idx]]
            tokens.append(token)        
            preds.append(label)
            truths.append(truth)
            idx +=1

    print(classification_report([truths], [preds], digits = 4, mode = 'strict'))
    print(classification_report(truths, preds, digits = 4))



# Author AIMPED
# Date 2023-March-14
# Description This file contains the pipeline for relation extraction.

import itertools
import pandas as pd
import numpy as np

def RelationAnnotateSentence(df):
    """It annotates the sentence with [Entity] tags.
    parameters:
    ----------------
    df: pandas dataframe
    return:
    ----------------
    df['new_sentence']: str
    """

    a1_start = min(df['sent_begin1'] - 1, df['sent_begin2'])
    a1_end = min(df['sent_end1'] + 1, df['sent_end2'] + 1)

    a2_start = max(df['sent_begin1'] - 1, df['sent_begin2'])
    a2_end = max(df['sent_end1'] + 1, df['sent_end2'] + 1)

    df['new_sentence'] = " ".join([
        df['sentence'][:a1_start],
        'e1b',
        df['sentence'][a1_start:a1_end],
        'e1e',
        df['sentence'][a1_end:a2_start],
        'e2b',
        df['sentence'][a2_start:a2_end],
        'e2e',
        df['sentence'][a2_end:]
    ])

    return df['new_sentence']


def RelationResults(sentences, ner_chunk_results, relation_classifier,
                    relation_white_label_list, relation_pairs, return_svg):
    """It returns the relation results of a text.
    parameters:
    ----------------
    sentences: list of str
    ner_chunk_results: list of dict
    relation_classifier: str
    ner_white_label_list: list of str
    relation_white_label_list: list of str
    relation_pairs: list of tuple
    return_svg: bool
    return:
    ----------------
    results: list of dict
    """

    end_df = pd.DataFrame()
    if len(ner_chunk_results) == 0:
        return end_df.to_dict(orient='records')
    else:
        df_ner_chunk_results = pd.DataFrame(ner_chunk_results)
        for i in df_ner_chunk_results.sent_idx.unique():
            sentence_bazli_ner_results = df_ner_chunk_results[df_ner_chunk_results.sent_idx == i]
            sentence_bazli_ner_results = sentence_bazli_ner_results.values.tolist()
            if len(sentence_bazli_ner_results) >= 2:
                df = pd.DataFrame(itertools.combinations(sentence_bazli_ner_results, 2))
                df['firstCharEnt1'] = df[0].apply(lambda x: x[2])
                df['lastCharEnt1'] = df[0].apply(lambda x: x[3])
                df['entity1'] = df[0].apply(lambda x: x[0])
                df['chunk1'] = df[0].apply(lambda x: x[1])
                df['sent_begin1'] = df[0].apply(lambda x: x[-2])
                df['sent_end1'] = df[0].apply(lambda x: x[-1])
                df['firstCharEnt2'] = df[1].apply(lambda x: x[2])
                df['lastCharEnt2'] = df[1].apply(lambda x: x[3])
                df['entity2'] = df[1].apply(lambda x: x[0])
                df['chunk2'] = df[1].apply(lambda x: x[1])
                df['sent_begin2'] = df[1].apply(lambda x: x[-2])
                df['sent_end2'] = df[1].apply(lambda x: x[-1])

                df = df[df.apply(lambda row: any((row['entity1'], row['entity2']) == item for item in relation_pairs), axis=1)]


                if len(df) != 0:
                    df['sentID'] = i
                    df['sentence'] = sentences[i]
                    df = df.drop([0, 1], axis=1)
                    df['new_sentence'] = np.nan
                    df['new_sentence'] = df.apply(RelationAnnotateSentence, axis=1)
                    df.reset_index(drop=True, inplace=True)
                    rel_results = relation_classifier(list(df['new_sentence']))
                    df = pd.concat([df, pd.DataFrame(rel_results)], axis=1)
                    df = df[['sentID', 'sentence', 'firstCharEnt1', 'sent_begin1', 'lastCharEnt1', 'sent_end1', 'entity1',
                             'chunk1',
                             'firstCharEnt2', 'sent_begin2', 'lastCharEnt2', 'sent_end2', 'entity2', 'chunk2', 'label',
                             'score'
                             ]]
                    df = df[df['label'].isin(relation_white_label_list)]
                    end_df = pd.concat([end_df, df], ignore_index=True)

        if end_df.empty or return_svg:
            return end_df.to_dict(orient='records')

        else:
            return end_df[['firstCharEnt1', 'lastCharEnt1', 'entity1', 'chunk1',
                           'firstCharEnt2', 'lastCharEnt2', 'entity2', 'chunk2', 'label', 'score']].to_dict(
                orient='records')

########################## neo4j knowledge graph ##########################
import json
from neo4j import GraphDatabase

## Creating a Connection Class
class Neo4j:

    def __init__(self, uri, user, pwd, db=None):

        self.__url = uri
        self.__user = user
        self.__pwd = pwd
        self.__driver = None
        self.db = db
        try:
            self.__driver = GraphDatabase.driver(self.__url, auth=(self.__user, self.__pwd))
            print("Connection Successful!")

        except Exception as e:
            print("Failed to create the driver:", e)

    def close(self):

        if self.__driver is not None:
            self.__driver.close()
            

    def create_neo4j_query(self, json_output=None, json_file=None):
        
        if json_output:
            data = json_output["output"]["data_json"]["result"]
        elif json_file:
            with open(json_file) as f:
                data = json.load(f)
            data = data["output"]["data_json"]["result"]
            
        queries = []
        for i in data:
            for item in i:
                entity1 = item["entity1"].replace("-","_")
                chunk1 = item["chunk1"].replace("'","")
                entity2 = item["entity2"].replace("-","_")
                chunk2 = item["chunk2"].replace("-","_")
                label = item["label"].replace("-","_")
                
                query = f"""
                MERGE (e1:{entity1} {{name: '{chunk1}'}})
                MERGE (e2:{entity2} {{name: '{chunk2}'}})
                MERGE (e1)-[:{label}]->(e2)
                """
                queries.append(query)
        print("Queries Created!")
        return queries            

    def run_query(self, query, parameters=None, db=None):
        if db is None:
            db = self.db
        assert self.__driver is not None, "Driver not initialized!"
        session = None
        response = None

        try:
            session = self.__driver.session(database=db) if db is not None else self.__driver.session()
            response = list(session.run(query, parameters))
        except Exception as e:
            print("Query failed:", e)
        finally:
            if session is not None:
                session.close()
        return response
    
    def write_data(self, queries=None, db=None):
        if db is None:
            db = self.db
        try:
            assert self.__driver is not None, "Driver not initialized!"
        except AssertionError as e:
            print(e)
        try:
            assert isinstance(queries, list), "Queries should be a list!"
        except AssertionError as e:
            print(e)
        try:
            assert len(queries) > 0, "Queries list is empty!"
        except AssertionError as e:
            print(e)
        try:
            print("Writing data to Neo4j...")
            for query in queries:
                self.run_query(query=query, db=db)  
            print("Data written successfully!")  
        except Exception as e:
            print("Failed to write data:", e)

    
    






    
